/**
 * Copyright (c) 2021 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "poker.h"

void pokerInit(poker2_t *poker) {
  poker->playerCount = 0;
  poker->playerDealer = 0;
  poker->playerSmallBlind = 0;
  poker->playerBigBlind = 0;
  pokerResetRound(poker);
}

void pokerResetRound(poker2_t *poker) {
  uint8_t i;
  poker2player_t *player;

  poker->deckSize = cardWriteDeck(poker->deck);
  poker->graveSize = 0;
  poker->communitySize = 0;
  poker->potCount = 0;
  pokerPotAdd(poker);

  for(i = 0; i < poker->playerCount; i++) {
    player = poker->players + i;
    player->cardCount = 0;
    player->currentBet = 0;
    player->state &= ~(
      POKER_PLAYER_STATE_FOLDED |
      POKER_PLAYER_STATE_HAS_BET_THIS_ROUND
    );
  }
}

void pokerNewDealer(poker2_t *poker) {
  uint8_t i, j, k;
  poker2player_t *player;
  bool foundDealer;
  bool foundSmall;

  foundDealer = false;
  foundSmall = false;

  j = poker->playerDealer + 1;
  for(i = 0; i < poker->playerCount; i++) {
    k = (j + i) % poker->playerCount;
    player = poker->players + k;
    if(player->state & POKER_PLAYER_STATE_OUT) continue;
    if(!foundDealer) {
      poker->playerDealer = k;
      foundDealer = true;
    } else if(!foundSmall) {
      poker->playerSmallBlind = k;
      foundSmall = true;
    } else {
      poker->playerBigBlind = k;
      break;
    }
  }
}

void pokerTakeBlinds(poker2_t *poker, int32_t small, int32_t big) {
  pokerPlayerBet(poker, poker->playerSmallBlind, small);
  pokerPlayerBet(poker, poker->playerBigBlind, big);
}

int32_t pokerGetCallValue(poker2_t *poker) {
  uint8_t i;
  int32_t call;
  call = 0;
  for(i = 0; i < poker->playerCount; i++) {
    call = mathMax(call, poker->players[i].currentBet);
  }
  return call;
}

// Pot functions
uint8_t pokerPotAdd(poker2_t *poker) {
  poker2pot_t *pot;
  uint8_t i = poker->potCount++;
  pot = poker->pots + i;
  pot->chips = 0;
  pot->playerCount = 0;
  return i;
}

void pokerPotAddPlayer(poker2pot_t *pot, uint8_t playerIndex) {
  if(arrayContains(
    sizeof(uint8_t), pot->players, pot->playerCount, &playerIndex
  )) return;
  pot->players[pot->playerCount++] = playerIndex;
}

// Dealer Functions
void pokerTurn(poker2_t *poker, uint8_t count) {
  uint8_t i;
  for(i = 0; i < count; i++) {
    cardDeal(
      poker->deck, &poker->deckSize, poker->community, &poker->communitySize
    );
  }
}

void pokerBurn(poker2_t *poker, uint8_t count) {
  uint8_t i;
  for(i = 0; i < count; i++) {
    cardDeal(poker->deck, &poker->deckSize, poker->grave, &poker->graveSize);
  }
}

// Player Functions
uint8_t pokerPlayerAdd(poker2_t *poker) {
  poker2player_t *player;
  uint8_t i = poker->playerCount++;
  
  player = poker->players + i;
  player->cardCount = 0;
  player->chips = 0;
  player->currentBet = 0;
  player->state = POKER_PLAYER_STATE_OUT;
  
  return i;
}

void pokerPlayerDeal(poker2_t *poker, poker2player_t *player, uint8_t count) {
  uint8_t i;
  for(i = 0; i < count; i++) {
    cardDeal(poker->deck, &poker->deckSize, player->cards, &player->cardCount);
  }
}

void pokerPlayerChipsAdd(poker2player_t *player, int32_t chips) {
  player->chips += chips;
  if(player->chips > 0) {
    player->state = flagOff(player->state, POKER_PLAYER_STATE_OUT);
  }
}

void pokerPlayerDealAll(poker2_t *poker, uint8_t count) {
  uint8_t i, j;
  poker2player_t *player;

  for(j = 0; j < count; j++) {
    for(i = 0; i < poker->playerCount; i++) {
      player = poker->players + i;
      
      // Can't deal to a player who is folded or out
      if(player->state & (POKER_PLAYER_STATE_FOLDED|POKER_PLAYER_STATE_OUT)) {
        continue;
      }

      pokerPlayerDeal(poker, player, 1);
    }
  }
}

bool pokerPlayerDoesNeedToBetThisRound(poker2_t *poker, uint8_t playerIndex) {
  poker2player_t *player;
  player = poker->players + playerIndex;
  if(player->state & POKER_PLAYER_STATE_FOLDED) return false;
  if(!(player->state & POKER_PLAYER_STATE_HAS_BET_THIS_ROUND)) return true;
  if(player->currentBet < pokerGetCallValue(poker)) return true;
  return false;
}

uint8_t pokerPlayerGetRemainingBetter(poker2_t *poker) {
  uint8_t i, j;

  for(i = 0; i < poker->playerCount; i++) {
    j = (i + poker->playerBigBlind + 1) % poker->playerCount;
    if(pokerPlayerDoesNeedToBetThisRound(poker, j)) return j;
  }
  return 0xFF;
}

uint8_t pokerPlayerGetNextBetter(poker2_t *poker, uint8_t current) {
  uint8_t i, j;

  for(i = 0; i < poker->playerCount; i++) {
    j = (i + current + 1) % poker->playerCount;
    if(pokerPlayerDoesNeedToBetThisRound(poker, j)) return j;
  }
  return 0xFF;
}

int32_t pokerPlayerGetCallBet(poker2_t *poker, poker2player_t *player) {
  return pokerGetCallValue(poker) - player->currentBet;
}

// Betting
void pokerPlayerBetPot(
  poker2_t *poker, poker2pot_t *pot, uint8_t playerIndex, int32_t chips
) {
  poker2player_t *player;
  player = poker->players + playerIndex;
  player->chips -= chips;
  player->currentBet += chips;
  pot->chips += chips;
  player->state |= POKER_PLAYER_STATE_HAS_BET_THIS_ROUND;
  pokerPotAddPlayer(pot, playerIndex);
}

void pokerPlayerBet(poker2_t *poker, uint8_t playerIndex, int32_t chips) {
  pokerPlayerBetPot(
    poker, poker->pots + (poker->potCount - 1), playerIndex, chips
  );
}

bool pokerPlayerCanCheck(poker2_t *poker, poker2player_t *player) {
  return pokerGetCallValue(poker) <= player->currentBet;
}

poker2turn_t pokerTurnFold(poker2_t *poker, uint8_t player) {
  return (poker2turn_t){
    .chips = 0,
    .confidence = 1,
    .type = POKER_TURN_TYPE_FOLD
  };
}

poker2turn_t pokerTurnBet(poker2_t *poker, uint8_t playerIndex, int32_t chips) {
  poker2turn_t turn;
  poker2player_t *player;
  int32_t i;
  player = poker->players + playerIndex;
  turn.confidence = 1;

  if(chips == 0) {
    turn.type = POKER_TURN_TYPE_CHECK;
    turn.chips = 0;
  } else if(player->chips <= chips) {
    turn.chips = player->chips;
    turn.type = POKER_TURN_TYPE_ALL_IN;
  } else {
    turn.chips = chips;
    turn.type = POKER_TURN_TYPE_BET;
    i = pokerGetCallValue(poker);

    if(chips == (i - player->currentBet)) {
      turn.type = POKER_TURN_TYPE_CALL;
    }
  }

  return turn;
}

// Winning
void pokerHandGetFull(
  poker2_t *poker, poker2player_t *player, card_t cards[POKER_WINNING_FULL_SIZE]
) {
  uint8_t i;

  // Add the dealer hand
  for(i = 0; i < poker->communitySize; i++) {
    cards[i] = poker->community[i];
  }

  // Add the player hand
  for(i = 0; i < player->cardCount; i++) {
    cards[i+poker->communitySize] = player->cards[i];
  }
}

void debug_WinnerFillRemaining(poker2playerwinning_t *winning) {
  uint8_t i, highest, current;
  card_t highestCard, currentCard;

  // Set the kicker
  winning->kicker = 0xFF;

  // Fill the remaining cards
  while(winning->setSize < POKER_WINNING_SET_SIZE) {
    highest = 0xFF;

    for(i = 0; i < winning->fullSize; i++) {
      currentCard = winning->full[i];
      if(cardContains(winning->set, winning->setSize, currentCard) != -1) {
        continue;
      }

      if(highest == 0xFF) {
        highestCard = currentCard;
        highest = cardGetNumber(highestCard);
      } else {
        current = cardGetNumber(currentCard);
        if(current != CARD_ACE && current < highest) continue;
        highestCard = currentCard;
        highest = current;
      }
    }

    if(highest == 0xFF) break;
    winning->set[winning->setSize++] = highestCard;
  }
  cardHandSort(winning->set, winning->setSize);
}

void pokerWinnerGetForPlayer(
  poker2_t *poker, poker2player_t *player, poker2playerwinning_t *winning
) {
  uint8_t i, j, l;
  int32_t index;
  card_t card;
  uint8_t number, suit, pairCount;
  int32_t pairs[CARD_SUIT_COUNT];

  // Get the full poker hand (should be a 7 card hand, but MAY not be)
  winning->fullSize = poker->communitySize + player->cardCount;
  pokerHandGetFull(poker, player, winning->full);
  cardHandSort(winning->full, winning->fullSize);

  // Reset the winning status.
  winning->setSize = 0;
  
  //////////////////////// Now look for the winning set ////////////////////////

  // Royal / Straight Flush
  for(i = 0; i < winning->fullSize; i++) {
    card = winning->full[i];
    number = cardGetNumber(card);
    if(number < CARD_FIVE) continue;

    suit = cardGetSuit(card);
    winning->setSize = 1;

    // Now look for the matching cards (Reverse order to order from A to 10)
    for(j = 1; j <= 4; j++) {
      l = number == CARD_FIVE && j == 4 ? CARD_ACE : number - j;//Ace low.
      index = cardContains(winning->full, winning->fullSize, cardGet(l, suit));
      if(index == -1) break;
      winning->set[j] = winning->full[index];
      winning->setSize++;
    }

    // Check if has all necessary cards.
    if(winning->setSize < POKER_WINNING_SET_SIZE) continue;

    // Add self to array
    winning->set[0] = winning->full[i];
    winning->type = (
      number == CARD_ACE ? POKER_WINNING_TYPE_ROYAL_FLUSH : 
      POKER_WINNING_TYPE_STRAIGHT_FLUSH
    );
    return;
  }

  // Four of a kind.
  winning->setSize = 0;
  for(i = 0; i < winning->fullSize; i++) {
    card = winning->full[i];
    number = cardGetNumber(card);
    pairCount = cardCountPairs(winning->full, winning->fullSize, number, pairs);
    if(pairCount < CARD_SUIT_COUNT) continue;

    winning->setSize = pairCount;
    for(j = 0; j < pairCount; j++) winning->set[j] = winning->full[pairs[j]];
    winning->type = POKER_WINNING_TYPE_FOUR_OF_A_KIND;
    debug_WinnerFillRemaining(winning);
    return;
  }

  // Full House
  winning->setSize = 0;
  for(i = 0; i < winning->fullSize; i++) {
    // Check we haven't already added this card.
    card = winning->full[i];
    if(cardContains(winning->set, winning->setSize, card) != -1) continue;

    number = cardGetNumber(card);
    pairCount = cardCountPairs(winning->full, winning->fullSize, number, pairs);

    // Did we find either two pair or three pair?
    if(pairCount != 2 && pairCount != 3) continue;
    if(winning->setSize == 3) pairCount = 2;//Clamp to 5 max.

    // Copy found pairs.
    for(j = 0; j < pairCount; j++) {
      winning->set[winning->setSize + j] = winning->full[pairs[j]];
    }
    winning->setSize += pairCount;

    // Winned?
    if(winning->setSize != POKER_WINNING_SET_SIZE) continue;
    winning->type = POKER_WINNING_TYPE_FULL_HOUSE;
    printf("Full House\n");
    return;
  }

  // Flush (5 same suit)
  for(i = 0; i < winning->fullSize; i++) {
    card = winning->full[i];
    suit = cardGetSuit(card);
    winning->setSize = 1;
    for(j = i+1; j < winning->fullSize; j++) {
      if(cardGetSuit(winning->full[j]) != suit) continue;
      winning->set[winning->setSize] = winning->full[j];
      winning->setSize++;
      if(winning->setSize == POKER_WINNING_SET_SIZE) break;
    }
    if(winning->setSize < POKER_WINNING_SET_SIZE) continue;
    winning->set[0] = winning->full[0];
    winning->type = POKER_WINNING_TYPE_FLUSH;
    return;
  }

  // Straight (sequence any suit)
  winning->setSize = 0;
  for(i = 0; i < winning->fullSize; i++) {
    card = winning->full[i];
    number = cardGetNumber(card);
    if(number < CARD_FIVE) continue;
    winning->setSize = 1;
    
    for(j = 1; j <= 4; j++) {
      l = number == CARD_FIVE && j == 4 ? CARD_ACE : number - j;//Ace low.
      index = cardContainsNumber(winning->full, winning->fullSize, l);
      if(index == -1) break;
      winning->set[j] = winning->full[index];
      winning->setSize++;
    }

    // Check if has all necessary cards.
    if(winning->setSize < POKER_WINNING_SET_SIZE) continue;
    winning->set[0] = winning->full[i];
    winning->type = POKER_WINNING_TYPE_STRAIGHT;
    return;
  }

  // Three of a kind
  winning->setSize = 0;
  for(i = 0; i < winning->fullSize; i++) {
    card = winning->full[i];
    number = cardGetNumber(card);
    pairCount = cardCountPairs(winning->full, winning->fullSize, number, pairs);
    if(pairCount != 3) continue;

    winning->setSize = pairCount;
    for(j = 0; j < pairCount; j++) winning->set[j] = winning->full[pairs[j]];
    winning->type = POKER_WINNING_TYPE_THREE_OF_A_KIND;
    debug_WinnerFillRemaining(winning);
    return;
  }

  // Two Pair
  winning->setSize = 0;
  for(i = 0; i < winning->fullSize; i++) {
    card = winning->full[i];// Check we haven't already added this card.
    if(cardContains(winning->set, winning->setSize, card) != -1) {
      continue;
    }

    number = cardGetNumber(card);
    pairCount = cardCountPairs(winning->full, winning->fullSize, number, pairs);
    if(pairCount != 2) continue;
    
    for(j = 0; j < pairCount; j++) {
      winning->set[j] = winning->full[winning->setSize + pairs[j]];
    }
    // arrayCopy(sizeof(int32_t), pairs, pairCount, winning->set+winning->setSize);
    winning->setSize += pairCount;
    if(winning->setSize != 4) continue;

    winning->type = POKER_WINNING_TYPE_TWO_PAIR;
    debug_WinnerFillRemaining(winning);
    return;
  }

  // Pair
  if(winning->setSize == 2) {
    winning->type = POKER_WINNING_TYPE_PAIR;
    debug_WinnerFillRemaining(winning);
    return;
  }

  // High card
  winning->setSize = 0;
  debug_WinnerFillRemaining(winning);
  winning->type = POKER_WINNING_TYPE_HIGH_CARD;

  return;
}